from __future__ import division
import math

import pygame
from pygame.draw import circle, line, rect

import data


class Sprite(object):

    def __init__(self, surfaces, frequency=5):
        self.surfaces = surfaces
        self.active_surfaces = surfaces[0]
        self.surface_id = 0
        self.surface = self.active_surfaces[self.surface_id]
        self.tts = self.frequency = frequency

    def select(self, index):
        self.active_surfaces = self.surfaces[index]
        self.surface_id = self.surface_id % len(self.active_surfaces)
        self.surface = self.active_surfaces[self.surface_id]

    def update(self):
        self.tts -= 1
        if self.tts <= 0:
            self.tts = self.frequency
            self.surface_id = (self.surface_id + 1) % len(self.active_surfaces)
            self.surface = self.active_surfaces[self.surface_id]


class Peg(object):

    "A screen coordinate serving as an endpoint for Links."

    def __init__(self, pos, color=None):
        self.x, self.y = self.pos = pos
        self.color = color
        self.links = set()
        self.rect = pygame.Rect(pos, (19, 19))
        self.rect.move_ip(-9, -9)

    def distance(self, other):
        dx = other.x - self.x
        dy = other.y - self.y
        return int((dx * dx + dy * dy) ** 0.5)

    def draw(self, playing_area, highlight=False):
        surface = playing_area.surface
        if highlight:
            light_color = (0, 0xff, 0)
            color = (0, 0xbf, 0)
            radius = 9
        else:
            light_color = (0xbf, 0xbf, 0xbf)
            color = (0x7f, 0x7f, 0x7f)
            radius = 7
        circle(surface, color, self.pos, radius, 0)
        circle(surface, light_color, self.pos, radius, 1)


class String(object):

    "A resource used to create Links between Pegs."

    def __init__(self, length, base_capacity, count, color=None):
        self.length = length
        self.base_capacity = base_capacity
        self.count = count
        self.color = color
        self.link = None

    def get_capacity(self, peg1, peg2, length):
        if self.color and self.color == peg1.color == peg2.color:
            return self.base_capacity * 3
        return self.base_capacity


class Scissors(object):
    pass


class Link(object):

    "A location for Ants."

    def __init__(self, peg1, peg2):
        self.is_alive = True
        self.is_goal = False
        self.pegs = (peg1, peg2)
        self.ants = set()
        for peg in self.pegs:
            peg.links.add(self)
        self.length = peg1.distance(peg2)
        self.rect = None

    def break_link(self, channel=None):
        for peg in self.pegs:
            peg.links.remove(self)
        for ant in list(self.ants):
            ant.fall(channel)

    def position(self, offset):
        x = [peg.x for peg in self.pegs]
        y = [peg.y for peg in self.pegs]
        dx = x[1] - x[0]
        dy = y[1] - y[0]
        return (x[0] + dx * offset // self.length, y[0] + dy * offset // self.length)

    def other_peg(self, peg):
        if self.pegs[0] is peg:
            return self.pegs[1]
        elif self.pegs[1] is peg:
            return self.pegs[0]

    def direction(self, forward_peg):
        if self.pegs[0] is forward_peg:
            return -1
        elif self.pegs[1] is forward_peg:
            return 1

    def angle(self, forward_peg):
        dx = self.pegs[1].x - self.pegs[0].x
        dy = self.pegs[0].y - self.pegs[1].y
        angle = int(round(math.acos(dx / self.length) * 6 / math.pi))
        if dy < 0:
            angle = (12 - angle) % 12
        if self.pegs[0] is forward_peg:
            return (angle + 6) % 12
        elif self.pegs[1] is forward_peg:
            return angle

    def offset_from_peg(self, offset, peg):
        if self.pegs[0] is peg:
            return offset
        elif self.pegs[1] is peg:
            return self.length - offset

    def add_ant(self, ant):
        self.ants.add(ant)

    def remove_ant(self, ant):
        self.ants.remove(ant)

    def remove(self, state):
        state.links.remove(self)

    # Placeholders for StringLink
    def act(self, state):
        pass

    def draw(self, playing_area):
        pass


class StringLink(Link):

    "A breakable Link made from string."

    fade_time = 8

    def __init__(self, peg1, peg2, string):
        assert string.link is None, "tried to link a previously linked string"
        super(StringLink, self).__init__(peg1, peg2)
        self.string = string
        string.link = self
        self.capacity = string.get_capacity(peg1, peg2, self.length)
        self._load = self._old_load = 0
        self._fade_timer = 0
        self.rect = peg1.rect.union(peg2.rect)

    def break_link(self, channel=None):
        super(StringLink, self).break_link(channel)
        self.string.link = None
        self.is_alive = False

    def act(self, state):
        if self._fade_timer > 0:
            self._fade_timer -= 1
            state.dirty.append(self.rect)

    def _apparent_load(self):
        return (self._load + (self._old_load - self._load) * self._fade_timer
                / self.fade_time)

    def add_ant(self, ant):
        super(StringLink, self).add_ant(ant)
        self._old_load = int(round(self._apparent_load()))
        self._fade_timer = self.fade_time
        self._load += ant.weight
        if self._load > self.capacity:
            sound = data.get_sound("snap")
            channel = None
            if sound:
                channel = sound.play()
            self.break_link(channel)

    def remove_ant(self, ant):
        super(StringLink, self).remove_ant(ant)
        self._old_load = int(round(self._apparent_load()))
        self._fade_timer = self.fade_time
        self._load -= ant.weight

    def draw(self, playing_area):
        if self.is_alive:
            surface = playing_area.surface
            red = int(round(420 * self._apparent_load() / self.capacity))
            green = 420 - red
            if red > 255:
                red = 255
            if green > 165:
                green = 165
            blue = 420 - red - green
            color = (red, green, blue)
            line(surface, color, self.pegs[0].pos, self.pegs[1].pos, 10)


class Ant(object):

    def __init__(self, link, offset, next_peg, weight, speed, delay,
                 pathfinder, sprite):
        self.link = None
        self.weight = weight
        self.speed = speed
        self.tta = self.delay = delay
        self.pathfinder = pathfinder
        self.sprite = sprite
        self.is_alive = True
        self.rect = None
        self.move(link, offset, next_peg=next_peg)
        self.reposition()

    def act(self, state):
        self.sprite.update()
        self.tta -= 1
        if self.tta <= 0:
            self.tta = self.delay
            if self.link and self.link.is_alive:
                if self.current_speed:
                    state.dirty.append(self.rect)
                    self.offset += self.current_speed
                    if self.offset >= self.link.length:
                        self.update_path(self.offset - self.link.length)
                    elif self.offset <= 0:
                        self.update_path(0 - self.offset)
                    self.reposition()
                    state.dirty.append(self.rect)
                else:
                    self.set_path()
            else:
                self.fall()

    def move(self, link, offset=0, offset_from=None, next_peg=None):
        if self.link:
            self.link.remove_ant(self)
        self.link = link
        if link:
            if link.is_alive:
                if offset_from:
                    self.offset = link.offset_from_peg(offset, offset_from)
                else:
                    self.offset = offset
                if next_peg:
                    self.next_peg = next_peg
                else:
                    self.next_peg = link.pegs[1]
                self.current_speed = self.speed * link.direction(self.next_peg)
                self.sprite.select(link.angle(self.next_peg))
                link.add_ant(self)
                if self.link:
                    self.set_path()
            else:
                self.fall()

    def set_path(self):
        self.next_link = self.pathfinder.follow_path(self.link, self.next_peg)

    def update_path(self, offset):
        self.set_path()
        if self.next_link:
            self.move(self.next_link, offset, self.next_peg,
                      self.next_link.other_peg(self.next_peg))
        elif self.link.is_goal:
            self.exit()
        else:
            self.fall()

    def reposition(self):
        if self.link:
            w, h = self.sprite.surface.get_size()
            x, y = self.link.position(self.offset)
            self.rect = self.sprite.surface.get_rect()
            self.rect.move_ip(x - w // 2, y - h // 2)

    def exit(self):
        self.move(None)
        self.is_alive = False

    def fall(self, channel=None):
        self.move(None)
        self.is_alive = False
        sound = data.get_sound("splash")
        if sound:
            if channel:
                channel.queue(sound)
            else:
                sound.play()

    def remove(self, state):
        state.ants.remove(self)

    def draw(self, playing_area):
        surface = playing_area.surface
        surface.blit(self.sprite.surface, self.rect)


class AntFactory(object):

    rect = None

    def __init__(self, link, offset, next_peg, weight, speed, delay,
                 pathfinder_class, surfaces, count, ant_class=Ant, frequency=40,
                 initial_pause=100):
        self.link = link
        self.offset = offset
        self.next_peg = next_peg
        self.weight = weight
        self.speed = speed
        self.delay = delay
        self.pathfinder_class = pathfinder_class
        self.surfaces = surfaces
        self.count = count
        self.ant_class = ant_class
        self.frequency = frequency
        self.tta = frequency + initial_pause
        self.is_alive = True

    def act(self, state):
        if self.count > 0:
            self.tta -= 1
            if self.tta <= 0:
                self.tta = self.frequency
                ant = self.ant_class(self.link, self.offset, self.next_peg,
                                     self.weight, self.speed, self.delay,
                                     self.pathfinder_class(),
                                     Sprite(self.surfaces))
                state.ants.append(ant)
                state.dirty.append(ant.rect)
                self.count -= 1
        else:
            self.is_alive = False

    def remove(self, state):
        state.ant_factories.remove(self)
