from __future__ import with_statement
from contextlib import contextmanager

import pygame

import data
from .objects import String, StringLink, Peg


class PlayingArea(object):

    def __init__(self, surface):
        self.surface = surface
        self.background = data.load_background()

    def draw(self):
        self.surface.blit(self.background, self.surface.get_rect())


class Toolbar(object):

    def __init__(self, surface):
        self.surface = surface
        self.tools = [None] * 9
        self.selected_tool = None

    def draw(self):
        light = (0xdf, 0xdf, 0xdf)
        medium = (0xbf, 0xbf, 0xbf)
        dark = (0x7f, 0x7f, 0x7f)
        self.surface.fill(medium)
        for x, tool in zip(xrange(15, 798, 87), self.tools):
            pygame.draw.lines(self.surface, dark, False,
                              [(x, 12), (x, 86), (x+74, 86)], 2)
            pygame.draw.lines(self.surface, light, False,
                              [(x, 12), (x+74, 12), (x+74, 86)], 2)


class MouseController(object):

    def __init__(self, state):
        self.state = state
        self.drawing_rect = None
        self.highlight_rect = None
        self.start_pos = None

    def evt_active(self, event):
        if not event.gain:
            self.clear_drawing_line()

    def evt_mousebuttonup(self, event):
        if event.button == 1 and self.start_pos:
            start_peg = self.get_peg(self.start_pos)
            end_peg = self.get_peg(event.pos)
            self.start_pos = None
            if start_peg and end_peg:
                link = StringLink(start_peg, end_peg, String(0, 5, 0))
                self.state.links.append(link)
                self.state.dirty.append(link.rect)
        elif event.button == 3:
            peg = Peg(event.pos)
            self.state.pegs.append(peg)
            self.state.dirty.append(peg.rect)

    def evt_mousebuttondown(self, event):
        if event.button == 1:
            self.start_pos = event.pos

    def do_drawing(self, playing_area):
        surface = playing_area.surface
        self.state.dirty.extend([self.drawing_rect, self.highlight_rect])
        self.drawing_rect = self.highlight_rect = None
        start_peg = end_peg = None
        if self.start_pos:
            start_peg = self.get_peg(self.start_pos)
            if start_peg:
                self.start_pos = start_peg.pos
            else:
                self.start_pos = None
        end_pos = pygame.mouse.get_pos()
        if end_pos:
            end_peg = self.get_peg(end_pos)
            if end_peg:
                end_pos = end_peg.pos
        if self.start_pos and end_pos:
            blue = (0, 0, 0xff)
            self.drawing_rect = pygame.draw.line(surface, blue, self.start_pos,
                                                 end_pos, 10)
            self.drawing_rect.inflate_ip(20, 20)
            self.state.dirty.append(self.drawing_rect)
        # Redraw start and end pegs to put them above the line.
        if start_peg:
            start_peg.draw(playing_area)
        if end_peg:
            end_peg.draw(playing_area, highlight=True)
            self.highlight_rect = end_peg.rect
            self.state.dirty.append(self.highlight_rect)

    def clear_drawing_line(self):
        self.start_pos = None

    def get_peg(self, pos):
        for peg in self.state.pegs:
            if peg.rect.collidepoint(pos):
                return peg


@contextmanager
def clipping(surface, rect):
    surface.set_clip(rect)
    try:
        yield surface
    finally:
        surface.set_clip(None)
